## @file
# Module that encodes and decodes a EFI_CAPSULE_HEADER with a payload
#
# Copyright (c) 2018, Intel Corporation. All rights reserved.<BR>
# SPDX-License-Identifier: BSD-2-Clause-Patent
#
"""Module for encoding and decoding a EFI_CAPSULE_HEADER with a payload."""

import struct
import uuid

from edk2toollib.uefi.fmp_capsule_header import FmpCapsuleHeaderClass


class UefiCapsuleHeaderClass (object):
    r"""An object representing a UEFI_CAPSULE_HEADER.

    Attributes:
        CapsuleGuid (uuid.UUID):    6DCBD5ED-E82D-4C44-BDA1-7194199AD92A
        HeaderSize (int):           The size of the capsule header. This may be larger than the size of the
                                    EFI_CAPSULE_HEADER since CapsuleGuid may imply extended header entries
        OemFlags (int):             Bit-mapped list describing the capsule attributes. The Flag values of 0x0000 -
                                    0xFFFF are defined by CapsuleGuid. Flag values of 0x10000 - 0xFFFFFFFF are defined
                                    by this specification
        PersistAcrossReset (bool):  Flag pulled from OemFlags
        PopulateSystemTable (bool): Flag pulled from OemFlags
        InitiateReset (bool):       Flag pulled from OemFlags
        Payload (str):              string representing packed data as bytes (i.e. b'\x01\x00\x03')
        FmpCapsuleHeader (FmpCapsuleHeaderClass): Fmp Capsule Header

    ```
    typedef struct {
        EFI_GUID          CapsuleGuid;
        UINT32            HeaderSize;
        UINT32            Flags;
        UINT32            CapsuleImageSize;
    } EFI_CAPSULE_HEADER;

    #define CAPSULE_FLAGS_PERSIST_ACROSS_RESET          0x00010000
    #define CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE         0x00020000
    #define CAPSULE_FLAGS_INITIATE_RESET                0x00040000
    ```
    """
    _StructFormat = '<16sIIII'
    _StructSize = struct.calcsize(_StructFormat)

    EFI_FIRMWARE_MANAGEMENT_CAPSULE_ID_GUID = uuid.UUID('6DCBD5ED-E82D-4C44-BDA1-7194199AD92A')

    _CAPSULE_FLAGS_PERSIST_ACROSS_RESET = 0x00010000
    _CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE = 0x00020000
    _CAPSULE_FLAGS_INITIATE_RESET = 0x00040000

    def __init__(self):
        """Inits an empty object."""
        self.CapsuleGuid = self.EFI_FIRMWARE_MANAGEMENT_CAPSULE_ID_GUID
        self.HeaderSize = self._StructSize
        self.OemFlags = 0x0000
        self.PersistAcrossReset = False
        self.PopulateSystemTable = False
        self.InitiateReset = False
        self.CapsuleImageSize = self.HeaderSize
        self.Payload = b''
        self.FmpCapsuleHeader = None

    def Encode(self):
        r"""Serializes the Header + payload.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        Flags = self.OemFlags
        if self.PersistAcrossReset:
            Flags = Flags | self._CAPSULE_FLAGS_PERSIST_ACROSS_RESET
        if self.PopulateSystemTable:
            Flags = Flags | self._CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE
        if self.InitiateReset:
            Flags = Flags | self._CAPSULE_FLAGS_INITIATE_RESET

        # If we have an FmpCapsuleHeader, let's collapse that now.
        if self.FmpCapsuleHeader is not None:
            self.Payload = self.FmpCapsuleHeader.Encode()

        self.CapsuleImageSize = self.HeaderSize + len(self.Payload)

        UefiCapsuleHeader = struct.pack(
            self._StructFormat,
            self.CapsuleGuid.bytes_le,
            self.HeaderSize,
            Flags,
            self.CapsuleImageSize,
            0
        )

        return UefiCapsuleHeader + self.Payload

    def Decode(self, Buffer):
        """Loads data into the Object by parsing a buffer.

        Args:
            Buffer (obj): Buffer containing the data

        Returns:
            (str): string of binary representing the payload

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid Signature
            (ValueError): Invalid Header size
        """
        if len(Buffer) < self._StructSize:
            raise ValueError
        (CapsuleGuid, HeaderSize, Flags, CapsuleImageSize, Reserved) = struct.unpack(
            self._StructFormat,
            Buffer[0:self._StructSize]
        )
        if HeaderSize < self._StructSize:
            raise ValueError
        if CapsuleImageSize != len(Buffer):
            raise ValueError
        self.CapsuleGuid = uuid.UUID(bytes_le=CapsuleGuid)
        self.HeaderSize = HeaderSize
        self.OemFlags = Flags & 0xffff
        self.PersistAcrossReset = (Flags & self._CAPSULE_FLAGS_PERSIST_ACROSS_RESET) != 0
        self.PopulateSystemTable = (Flags & self._CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE) != 0
        self.InitiateReset = (Flags & self._CAPSULE_FLAGS_INITIATE_RESET) != 0
        self.CapsuleImageSize = CapsuleImageSize
        self.Payload = Buffer[self.HeaderSize:]
        if len(self.Payload) > 0 and self.CapsuleGuid == self.EFI_FIRMWARE_MANAGEMENT_CAPSULE_ID_GUID:
            self.FmpCapsuleHeader = FmpCapsuleHeaderClass()
            self.FmpCapsuleHeader.Decode(self.Payload)

        return self.Payload

    def DumpInfo(self):
        """Prints payload header information."""
        Flags = self.OemFlags
        if self.PersistAcrossReset:
            Flags = Flags | self._CAPSULE_FLAGS_PERSIST_ACROSS_RESET
        if self.PopulateSystemTable:
            Flags = Flags | self._CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE
        if self.InitiateReset:
            Flags = Flags | self._CAPSULE_FLAGS_INITIATE_RESET
        print('EFI_CAPSULE_HEADER.CapsuleGuid      = {Guid}'.format(Guid=str(self.CapsuleGuid).upper()))
        print('EFI_CAPSULE_HEADER.HeaderSize       = {Size:08X}'.format(Size=self.HeaderSize))
        print('EFI_CAPSULE_HEADER.Flags            = {Flags:08X}'.format(Flags=Flags))
        print('  OEM Flags                         = {Flags:04X}'.format(Flags=self.OemFlags))
        if self.PersistAcrossReset:
            print('  CAPSULE_FLAGS_PERSIST_ACROSS_RESET')
        if self.PopulateSystemTable:
            print('  CAPSULE_FLAGS_POPULATE_SYSTEM_TABLE')
        if self.InitiateReset:
            print('  CAPSULE_FLAGS_INITIATE_RESET')
        print('EFI_CAPSULE_HEADER.CapsuleImageSize = {Size:08X}'.format(Size=self.CapsuleImageSize))
        print('sizeof (Payload)                    = {Size:08X}'.format(Size=len(self.Payload)))
        if self.FmpCapsuleHeader is not None:
            self.FmpCapsuleHeader.DumpInfo()



# @file dsc_parser.py
# Code to help parse DSC files
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##
"""Code to help parse DSC files."""
import os

from edk2toollib.uefi.edk2.parsers.base_parser import HashFileParser


class DscParser(HashFileParser):
    """Object representing a parsed DSC file with a capability to parse.

    Attributes:
        SixMods (list): list of X64 Modules (the line in the file)
        SixModsEnhanced (list): Better parsed (in a dict) list of X64 Modules
        ThreeMods (list): list of IA32 Modules (the line in the file)
        ThreeModsEnhanced (list): Better parsed (in a dict) list of IA32 Modules
        OtherMods (list): list of other Mods that are not IA32, X64 specific
        Libs (list): list of Libs (the line in the file)
        LibsEnhanced (list): Better parsed (in a dict) list of Libs
        LibraryClassToInstanceDict (dict): Key (Library class) Value (Instance)
        Pcds (list): List of Pcds
    """
    def __init__(self):
        """Init an empty Parser."""
        super(DscParser, self).__init__('DscParser')
        self.SixMods = []
        self.SixModsEnhanced = []
        self.ThreeMods = []
        self.ThreeModsEnhanced = []
        self.OtherMods = []
        self.Libs = []
        self.LibsEnhanced = []
        self.ParsingInBuildOption = 0
        self.LibraryClassToInstanceDict = {}
        self.Pcds = []
        self._no_fail_mode = False
        self._dsc_file_paths = set()  # This includes the full paths for every DSC that makes up the file

    def __ParseLine(self, Line, file_name=None, lineno=None):
        line_stripped = self.StripComment(Line).strip()
        if (len(line_stripped) < 1):
            return ("", [], None)

        line_resolved = self.ReplaceVariables(line_stripped)
        if (self.ProcessConditional(line_resolved)):
            # was a conditional
            # Other parser returns line_resolved, [].  Need to figure out which is right
            return ("", [], None)

        # not conditional keep processing

        # check if conditional is active
        if (not self.InActiveCode()):
            return ("", [], None)

        # check for include file and import lines from file
        if (line_resolved.strip().lower().startswith("!include")):
            # include line.
            tokens = line_resolved.split()
            include_file = tokens[1]
            sp = self.FindPath(include_file)
            if sp is None:
                raise FileNotFoundError(include_file)
            self.Logger.debug("Opening Include File %s" % sp)
            self._PushTargetFile(sp)
            lf = open(sp, "r")
            loc = lf.readlines()
            lf.close()
            return ("", loc, sp)

        # check for new section
        (IsNew, Section) = self.ParseNewSection(line_resolved)
        if (IsNew):
            self.CurrentSection = Section.upper()
            self.Logger.debug("New Section: %s" % self.CurrentSection)
            self.Logger.debug("FullSection: %s" % self.CurrentFullSection)
            return (line_resolved, [], None)

        # process line in x64 components
        if (self.CurrentFullSection.upper() == "COMPONENTS.X64"):
            if (self.ParsingInBuildOption > 0):
                if (".inf" in line_resolved.lower()):
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    self.Logger.debug("Found Library in a 64bit BuildOptions Section: %s" % p)
                elif "tokenspaceguid" in line_resolved.lower() and \
                        line_resolved.count('|') > 0 and line_resolved.count('.') > 0:
                    # should be a pcd statement
                    p = line_resolved.partition('|')
                    self.Pcds.append(p[0].strip())
                    self.Logger.debug("Found a Pcd in a 64bit Module Override section: %s" % p[0].strip())
            else:
                if (".inf" in line_resolved.lower()):
                    p = self.ParseInfPathMod(line_resolved)
                    self.SixMods.append(p)
                    if file_name is not None and lineno is not None:
                        self.SixModsEnhanced.append({'file': os.path.normpath(file_name), 'lineno': lineno, 'data': p})
                    self.Logger.debug("Found 64bit Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in ia32 components
        elif (self.CurrentFullSection.upper() == "COMPONENTS.IA32"):
            if (self.ParsingInBuildOption > 0):
                if (".inf" in line_resolved.lower()):
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    if file_name is not None and lineno is not None:
                        self.LibsEnhanced.append({'file': os.path.normpath(file_name), 'lineno': lineno, 'data': p})
                    self.Logger.debug("Found Library in a 32bit BuildOptions Section: %s" % p)
                elif "tokenspaceguid" in line_resolved.lower() and \
                        line_resolved.count('|') > 0 and line_resolved.count('.') > 0:
                    # should be a pcd statement
                    p = line_resolved.partition('|')
                    self.Pcds.append(p[0].strip())
                    self.Logger.debug("Found a Pcd in a 32bit Module Override section: %s" % p[0].strip())

            else:
                if (".inf" in line_resolved.lower()):
                    p = self.ParseInfPathMod(line_resolved)
                    self.ThreeMods.append(p)
                    if file_name is not None and lineno is not None:
                        self.ThreeModsEnhanced.append({'file': os.path.normpath(file_name),
                                                       'lineno': lineno, 'data': p})
                    self.Logger.debug("Found 32bit Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in other components
        elif ("COMPONENTS" in self.CurrentFullSection.upper()):
            if (self.ParsingInBuildOption > 0):
                if (".inf" in line_resolved.lower()):
                    p = self.ParseInfPathLib(line_resolved)
                    self.Libs.append(p)
                    self.Logger.debug("Found Library in a BuildOptions Section: %s" % p)
                elif "tokenspaceguid" in line_resolved.lower() and \
                        line_resolved.count('|') > 0 and line_resolved.count('.') > 0:
                    # should be a pcd statement
                    p = line_resolved.partition('|')
                    self.Pcds.append(p[0].strip())
                    self.Logger.debug("Found a Pcd in a Module Override section: %s" % p[0].strip())

            else:
                if (".inf" in line_resolved.lower()):
                    p = self.ParseInfPathMod(line_resolved)
                    self.OtherMods.append(p)
                    self.Logger.debug("Found Module: %s" % p)

            self.ParsingInBuildOption = self.ParsingInBuildOption + line_resolved.count("{")
            self.ParsingInBuildOption = self.ParsingInBuildOption - line_resolved.count("}")
            return (line_resolved, [], None)

        # process line in library class section (don't use full name)
        elif (self.CurrentSection.upper() == "LIBRARYCLASSES"):
            if (".inf" in line_resolved.lower()):
                p = self.ParseInfPathLib(line_resolved)
                self.Libs.append(p)
                self.Logger.debug("Found Library in Library Class Section: %s" % p)
            return (line_resolved, [], None)
        # process line in PCD section
        elif (self.CurrentSection.upper().startswith("PCDS")):
            if "tokenspaceguid" in line_resolved.lower() and \
                    line_resolved.count('|') > 0 and line_resolved.count('.') > 0:
                # should be a pcd statement
                p = line_resolved.partition('|')
                self.Pcds.append(p[0].strip())
                self.Logger.debug("Found a Pcd in a PCD section: %s" % p[0].strip())
            return (line_resolved, [], None)
        else:
            return (line_resolved, [], None)

    def __ParseDefineLine(self, Line):
        line_stripped = self.StripComment(Line).strip()
        if (len(line_stripped) < 1):
            return ("", [])

        # this line needs to be here to resolve any symbols inside the !include lines, if any
        line_resolved = self.ReplaceVariables(line_stripped)
        if (self.ProcessConditional(line_resolved)):
            # was a conditional
            # Other parser returns line_resolved, [].  Need to figure out which is right
            return ("", [])

        # not conditional keep processing

        # check if conditional is active
        if (not self.InActiveCode()):
            return ("", [])

        # check for include file and import lines from file
        if (line_resolved.strip().lower().startswith("!include")):
            # include line.
            tokens = line_resolved.split()
            include_file = tokens[1]
            self.Logger.debug("Opening Include File %s" % include_file)
            sp = self.FindPath(include_file)
            if sp is None:
                raise FileNotFoundError(include_file)
            self._PushTargetFile(sp)
            lf = open(sp, "r")
            loc = lf.readlines()
            lf.close()
            return ("", loc)

        # check for new section
        (IsNew, Section) = self.ParseNewSection(line_resolved)
        if (IsNew):
            self.CurrentSection = Section.upper()
            self.Logger.debug("New Section: %s" % self.CurrentSection)
            self.Logger.debug("FullSection: %s" % self.CurrentFullSection)
            return (line_resolved, [])

        # process line based on section we are in
        if (self.CurrentSection == "DEFINES") or (self.CurrentSection == "BUILDOPTIONS"):
            if line_resolved.count("=") >= 1:
                tokens = line_resolved.split("=", 1)
                leftside = tokens[0].split()
                if (len(leftside) == 2):
                    left = leftside[1]
                else:
                    left = leftside[0]
                right = tokens[1].strip()

                self.LocalVars[left] = right
                self.Logger.debug("Key,values found:  %s = %s" % (left, right))

                # iterate through the existed LocalVars and try to resolve the symbols
                for var in self.LocalVars:
                    self.LocalVars[var] = self.ReplaceVariables(self.LocalVars[var])
                return (line_resolved, [])
        else:
            return (line_resolved, [])

    def ParseInfPathLib(self, line):
        """Parses a line with an INF path Lib."""
        if (line.count("|") > 0):
            line_parts = []
            c = line.split("|")[0].strip()
            i = line.split("|")[1].strip()
            if (c in self.LibraryClassToInstanceDict):
                line_parts = self.LibraryClassToInstanceDict.get(c)
            sp = self.FindPath(i)
            line_parts.append(sp)
            self.LibraryClassToInstanceDict[c] = line_parts
            return line.split("|")[1].strip()
        else:
            return line.strip().split()[0]

    def ParseInfPathMod(self, line):
        """Parses a line with an INF path."""
        return line.strip().split()[0].rstrip("{")

    def __ProcessMore(self, lines, file_name=None):
        """Runs after ProcessDefines and does a full parsing of the DSC.

        Everything is resolved to a final state
        """
        if (len(lines) == 0):
            return
        for index in range(0, len(lines)):
            # we try here so that we can catch exceptions from individual lines
            try:
                raw_line = lines[index]
                (line, add, new_file) = self.__ParseLine(raw_line, file_name=file_name, lineno=index + 1)
                if (len(line) > 0):
                    self.Lines.append(line)
                self.__ProcessMore(add, file_name=new_file)
            except Exception as e:
                # check if we're in no fail mode or not
                if not self._no_fail_mode:  # if we are, fail
                    raise
                else:
                    # otherwise, let the user know that we failed in the DSC
                    self.Logger.warning(f"DSC Parser (No-Fail Mode): {raw_line}")
                    self.Logger.warning(e)

    def __ProcessDefines(self, lines):
        """Goes through a file once to look for [Define] sections.

        Only Sections, DEFINE, X = Y, and !includes are resolved
        This resolves all the defines since they can be anywhere in a file.
        Ideally this should be run until we reach stable state but this parser is not
        accurate and is more of an approximation of what the real parser does.
        """
        if (len(lines) == 0):
            return
        for raw_line in lines:
            # we want to catch exceptions here since we are doing includes as we potentially might blow up
            # we want to catch on a line by line basis
            try:
                (line, add) = self.__ParseDefineLine(raw_line)
                self.__ProcessDefines(add)
            except Exception:
                # Since we're going to do this in ProcessMore, don't warn people if there's an exception
                # otherwise, raise the exception and act normally
                if not self._no_fail_mode:
                    raise

    def SetNoFailMode(self, enabled=True):
        """The parser won't throw exceptions when this is turned on.

        WARNING: This can result in some weird behavior
        """
        self._no_fail_mode = enabled

    def ResetParserState(self):
        """Resets the parser."""
        #
        # add more DSC parser based state reset here, if necessary
        #
        super(DscParser, self).ResetParserState()

    def ParseFile(self, filepath):
        """Parses the DSC file at the provided path."""
        self.Logger.debug("Parsing file: %s" % filepath)
        sp = self.FindPath(filepath)
        if sp is None:
            raise FileNotFoundError(filepath)
        self._PushTargetFile(sp)
        f = open(sp, "r")
        # expand all the lines and include other files
        file_lines = f.readlines()
        self.__ProcessDefines(file_lines)
        # reset the parser state before processing more
        self.ResetParserState()
        self._PushTargetFile(sp)
        self.__ProcessMore(file_lines, file_name=sp)
        f.close()
        self.Parsed = True

    def _PushTargetFile(self, targetFile):
        self.TargetFilePath = os.path.abspath(targetFile)
        self._dsc_file_paths.add(self.TargetFilePath)

    def GetMods(self):
        """Returns a list with all Mods."""
        return self.ThreeMods + self.SixMods

    def GetModsEnhanced(self):
        """Returns a list with all ModsEnhanced."""
        return self.ThreeModsEnhanced + self.SixModsEnhanced

    def GetLibs(self):
        """Returns a list with all Libs."""
        return self.Libs

    def GetLibsEnhanced(self):
        """Returns a list with all LibsEnhanced."""
        return self.LibsEnhanced

    def GetAllDscPaths(self):
        """Returns a list with all the paths that this DSC uses (the base file and any includes).

        They are not all guaranteed to be DSC files
        """
        return self._dsc_file_paths



# @file wincert.py
# Code to work with UEFI WinCert data
#
# Copyright (c) Microsoft Corporation
#
# SPDX-License-Identifier: BSD-2-Clause-Patent
##

"""Module for working with UEFI WinCert data."""


import io
import struct
import sys
import uuid
from warnings import warn

from pyasn1.codec.der.decoder import decode as der_decode
from pyasn1_modules import rfc2315

from edk2toollib.utility_functions import hexdump


class WinCertPkcs1(object):
    """Object representing a WinCertPkcs1 struct.

    Certificate which encapsulates the RSASSA_PKCS1-v1_5 digital signature.

    Attributes:
        Hdr_dwLength (int):         The length of the entire certificate, including the length of the header, in bytes.
        Hdr_wRevision (int):        The certificate type. See WIN_CERT_TYPE_xxx for the UEFI certificate types. The UEFI
                                    specification reserves the range of certificate type values from 0x0EF0 to 0x0EFF.
        Hdr_wCertificateType (int): The following is the actual certificate. The format of the certificate depends on
                                    wCertificateType.
        hash_algorithm (uuid.UUID):  The Guid representing the hash algorithm for the Cert.
        cert_data (memoryview):      The actual Cert.

    The WIN_CERTIFICATE_UEFI_PKCS1_15 structure is derived from
    WIN_CERTIFICATE and encapsulate the information needed to
    implement the RSASSA-PKCS1-v1_5 digital signature algorithm as
    specified in RFC2437.
    typedef struct {
        WIN_CERTIFICATE Hdr;
        EFI_GUID        HashAlgorithm;
        UINT8 Signature[];
    } WIN_CERTIFICATE_EFI_PKCS1_15;

    The WIN_CERTIFICATE structure is part of the PE/COFF specification.
    typedef struct {
        UINT32  dwLength;
        UINT16  wRevision;
        UINT16  wCertificateType;
        UINT8 bCertificate[ANYSIZE_ARRAY];
    } WIN_CERTIFICATE;
    """
    STATIC_STRUCT_SIZE = (4 + 2 + 2 + 16)
    EFI_HASH_SHA256 = uuid.UUID("{51AA59DE-FDF2-4EA3-BC63-875FB7842EE9}")  # EFI_HASH_SHA256 guid defined by UEFI spec

    def __init__(self, filestream=None):
        """Inits the object."""
        if filestream is None:
            self.Hdr_dwLength = WinCertPkcs1.STATIC_STRUCT_SIZE
            self.Hdr_wRevision = WinCert.REVISION
            self.Hdr_wCertificateType = WinCert.WIN_CERT_TYPE_EFI_PKCS115
            self.hash_algorithm = None
            self.cert_data = None
        else:
            self.decode(filestream)

    @property
    def HashAlgorithm(self):
        """Returns the HashAlgorithm."""
        warn("HashAlgorithm is deprecated. Use hash_algorithm instead.", DeprecationWarning, 2)
        return self.hash_algorithm

    @HashAlgorithm.setter
    def HashAlgorithm(self, value):
        """Sets the HashAlgorithm."""
        warn("HashAlgorithm is deprecated. Use hash_algorithm instead.", DeprecationWarning, 2)
        self.hash_algorithm = value

    @property
    def CertData(self):
        """Returns the CertData."""
        warn("CertData is deprecated. Use cert_data instead.", DeprecationWarning, 2)
        return self.cert_data

    @CertData.setter
    def CertData(self, value):
        """Sets the CertData."""
        warn("CertData is deprecated. Use cert_data instead.", DeprecationWarning, 2)
        self.cert_data = value

    def add_cert_data(self, fs):
        """Adds the Cert Data to the struct.

        Args:
            fs (io.BytesIO): filestream representing bytes

        Raises:
            (ValueError): Invalid Cert Data
            (ValueError): Missing Hash Algorithm
        """
        if self.cert_data is not None:
            raise ValueError("Certificate data already set")

        if self.hash_algorithm is None:
            raise ValueError("You must set the Hash Algorithm first")

        self.cert_data = memoryview(fs.read())
        self.Hdr_dwLength = self.Hdr_dwLength + len(self.cert_data)

    def AddCertData(self, fs):
        """Adds the Cert Data to the struct.

        Raises:
            (ValueError): Invalid Cert Data
            (ValueError): Missing Hash Algorithm
        """
        warn("AddCertData is deprecated, use add_cert_data instead", DeprecationWarning, 2)
        self.add_cert_data(fs)

    def set_hash_algorithm(self, hash_algorithm: uuid.UUID):
        """Sets the hash algorithm for the wincert.

        Args:
            hash_algorithm (uuid.UUID): The Guid representing the hash algorithm for the Cert.
        """
        self.hash_algorithm = hash_algorithm

    def decode(self, fs):
        """Populates the struct from a filestream.

        Args:
            fs (obj): An open file

        Raises:
            (ValueError): Invalid stream
            (ValueError): Invalid stream size
        """
        if (fs is None):
            raise ValueError("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if ((end - offset) < WinCertPkcs1.STATIC_STRUCT_SIZE):  # size of the static header data
            raise ValueError("Invalid file stream size")

        self.Hdr_dwLength = struct.unpack("=I", fs.read(4))[0]
        self.Hdr_wRevision = struct.unpack("=H", fs.read(2))[0]
        self.Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]
        self.hash_algorithm = uuid.UUID(bytes_le=fs.read(16))
        self.cert_data = None

        if ((end - fs.tell()) < 1):
            raise ValueError("Invalid File stream. No data for signature cert data")

        if ((end - fs.tell()) < (self.Hdr_dwLength - WinCertPkcs1.STATIC_STRUCT_SIZE)):
            raise ValueError("Invalid file stream size")

        self.cert_data = memoryview(fs.read(self.Hdr_dwLength - WinCertPkcs1.STATIC_STRUCT_SIZE))

    def PopulateFromFileStream(self, fs):
        """Populates the struct from a filestream.

        Args:
            fs (obj): an open file

        Raises:
            (ValueError): Invalid stream
            (ValueError): Invalid stream size
        """
        warn("PopulateFromFileStream is deprecated, use decode instead", DeprecationWarning, 2)
        return self.decode(fs)

    def print(self, out_fs=sys.stdout):
        """Prints the struct to the console."""
        out_fs.write("\n-------------------- WinCertPKCS115 ---------------------\n")
        out_fs.write(f"  Hdr_dwLength:         0x{self.Hdr_dwLength:0X}\n")
        out_fs.write(f"  Hdr_wRevision:        0x{self.Hdr_wRevision:0X}\n")
        out_fs.write(f"  Hdr_wCertificateType: 0x{self.Hdr_wCertificateType:0X}\n")
        out_fs.write(f"  Hash Guid:            {str(self.hash_algorithm)}\n")
        out_fs.write("  cert_data:             \n")
        cdl = self.cert_data.tolist()
        hexdump(cdl, out_fs=out_fs)

    def Print(self, out_fs=sys.stdout):
        """Prints the struct to the console."""
        warn("Print is deprecated, use print instead", DeprecationWarning, 2)
        self.print(out_fs)

    def encode(self):
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        output = b""
        output += struct.pack("=I", self.Hdr_dwLength)
        output += struct.pack("=H", self.Hdr_wRevision)
        output += struct.pack("=H", self.Hdr_wCertificateType)
        output += self.hash_algorithm.bytes_le
        output += self.cert_data

        return output

    def write(self, fs):
        r"""Writes an serialized object to a filestream.

        Args:
            fs (obj): an open file

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        fs.write(self.encode())

    def Write(self, fs):
        r"""Serializes the object.

        Args:
            fs (obj): an open file

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        warn("Write is deprecated, use encode instead", DeprecationWarning, 2)
        self.write(fs)


class WinCertUefiGuid(object):
    """Object representing a Certificate which encapsulates a GUID-specific digital signature.

    Attributes:
        Hdr_dwLength (int):         The length of the entire certificate, including the length of the header, in bytes.
        Hdr_wRevision (int):        The certificate type. See WIN_CERT_TYPE_xxx for the UEFI certificate types. The UEFI
                                    specification reserves the range of certificate type values from 0x0EF0 to 0x0EFF.
        Hdr_wCertificateType (int): The following is the actual certificate. The format of the certificate depends on
                                    wCertificateType.
        cert_data (memoryview):      The actual Cert.

    typedef struct {
        WIN_CERTIFICATE   Hdr;
        EFI_GUID          cert_type;
        UINT8            cert_data[1];
    } WIN_CERTIFICATE_UEFI_GUID;

    The WIN_CERTIFICATE structure is part of the PE/COFF specification.
    typedef struct {
        UINT32  dwLength;
        UINT16  wRevision;
        UINT16  wCertificateType;
        UINT8 bCertificate[ANYSIZE_ARRAY];
    } WIN_CERTIFICATE;
    """

    _StructFormat = '<L2H16s'
    _StructSize = struct.calcsize(_StructFormat)

    _EFI_CERT_TYPE_PKCS7_GUID = uuid.UUID('4aafd29d-68df-49ee-8aa9-347d375665a7')

    # Preserved for back compat.
    STATIC_STRUCT_SIZE = _StructSize
    PKCS7Guid = _EFI_CERT_TYPE_PKCS7_GUID

    def __init__(self, in_data=None):
        """Inits the object."""
        self.Hdr_dwLength = self._StructSize
        self.Hdr_wRevision = WinCert.REVISION
        self.Hdr_wCertificateType = WinCert.WIN_CERT_TYPE_EFI_GUID
        self.cert_type = self._EFI_CERT_TYPE_PKCS7_GUID
        self.cert_data = b''

        if in_data is not None:
            self.decode(in_data)

    @property
    def CertData(self):
        """Returns the CertData."""
        warn("CertData is deprecated. Use cert_data instead.", DeprecationWarning, 2)
        return self.cert_data

    @CertData.setter
    def CertData(self, value):
        """Sets the CertData."""
        warn("CertData is deprecated. Use cert_data instead.", DeprecationWarning, 2)
        self.cert_data = value

    @property
    def CertType(self):
        """Returns the CertType."""
        warn("CertType is deprecated. Use cert_type instead.", DeprecationWarning, 2)
        return self.cert_type

    @CertType.setter
    def CertType(self, value):
        """Sets the CertType."""
        warn("CertType is deprecated. Use cert_type instead.", DeprecationWarning, 2)
        self.cert_type = value

    def get_length(self):
        """Returns the length of the WinCertUefiGuid and it's data."""
        return self.Hdr_dwLength

    def Encode(self):
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')

        Raises:
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        warn("Encode is deprecated, use encode instead", DeprecationWarning, 2)
        return self.encode()

    def encode(self):
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')

        Raises:
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        if self.Hdr_wRevision != WinCert.REVISION:
            raise ValueError("Invalid revision")
        if self.Hdr_wCertificateType != WinCert.WIN_CERT_TYPE_EFI_GUID:
            raise ValueError("Invalid win_certificate type (WIN_CERT_TYPE_EFI_GUID)")
        if self.cert_type != self._EFI_CERT_TYPE_PKCS7_GUID:
            raise ValueError("Invalid certificate type (EFI_CERT_TYPE_PKCS7_GUID)")
        self.Hdr_dwLength = self._StructSize + len(self.cert_data)

        win_cert_header = struct.pack(
            self._StructFormat,
            self.Hdr_dwLength,
            self.Hdr_wRevision,
            self.Hdr_wCertificateType,
            self.cert_type.bytes_le
        )

        return win_cert_header + self.cert_data

    def Decode(self, Buffer):
        """Loads the struct with values from a buffer.

        Args:
            Buffer (filestream | bytes): Buffer containing serialized data

        Returns:
            (obj): Any remaining buffer

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid dwlength
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        warn("Decode is deprecated, use decode instead", DeprecationWarning, 2)
        return self.decode(Buffer)

    def decode(self, in_data):
        """Loads the struct with values from a buffer.

        Args:
            in_data (filestream | bytes): in_data containing serialized data

        Returns:
            (obj): Any remaining buffer

        Raises:
            (ValueError): Invalid Buffer
            (ValueError): Invalid dwLength
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
            (ValueError): Invalid datatype provided

        """
        if hasattr(in_data, 'seek'):  # Filestream like object
            return self._from_filestream(in_data)
        elif hasattr(in_data, 'decode'):  # Bytes like object
            return self._from_buffer(in_data)
        else:
            raise ValueError(
                f"Invalid datatype provided: {type(in_data)}, data may only be of type filestream or bytes")

    def _from_buffer(self, buffer):
        """Loads the struct with values from a buffer.

        Args:
            buffer (obj): Buffer containing serialized data

        Returns:
            (obj): Any remaining buffer

        Raises:
            (ValueError): Invalid buffer
            (ValueError): Invalid dwLength
            (ValueError): Invalid Revision
            (ValueError): Invalid Cert Type
        """
        if len(buffer) < self._StructSize:
            raise ValueError
        (dwLength, wRevision, wCertificateType, cert_type) = struct.unpack(
            self._StructFormat,
            buffer[0:self._StructSize]
        )
        if dwLength < self._StructSize:
            raise ValueError
        if wRevision != WinCert.REVISION:
            raise ValueError
        if wCertificateType != WinCert.WIN_CERT_TYPE_EFI_GUID:
            raise ValueError
        if cert_type != self._EFI_CERT_TYPE_PKCS7_GUID.bytes_le:
            raise ValueError
        self.Hdr_dwLength = dwLength
        self.Hdr_wRevision = wRevision
        self.Hdr_wCertificateType = wCertificateType
        self.cert_type = uuid.UUID(bytes_le=cert_type)
        self.cert_data = buffer[self._StructSize:self.Hdr_dwLength]

        # Return the remaining buffer, if any exists.
        return buffer[self.Hdr_dwLength:]

    def _from_filestream(self, fs):
        """Un-serialized from a filestream.

        Args:
            fs (obj): Already opened file

        Raises:
            (ValueError): Invalid fs
            (ValueError): Invalid size
        """
        if fs is None:
            raise ValueError

        # Determine the end of the stream.
        current = fs.tell()
        end = fs.seek(0, io.SEEK_END)
        fs.seek(current)

        # Make sure that we can at least parse the size field.
        field_string = "<I"
        field_size = struct.calcsize(field_string)
        if (end - current) < field_size:
            raise ValueError

        # Parse the size field.
        (buffer_size,) = struct.unpack(field_string, fs.read(field_size))

        if (end - current) < buffer_size:
            raise ValueError

        fs.seek(current)
        object_buffer = fs.read(buffer_size)

        return self._from_buffer(object_buffer)

    def AddCertData(self, in_data):
        """Adds the Cert Data to the struct.

        Args:
            in_data (obj): Data to read.
        """
        warn("AddCertData is deprecated, use add_cert_data instead", DeprecationWarning, 2)
        return self.add_cert_data(in_data)

    def add_cert_data(self, in_data):
        """Adds the Cert Data to the struct.

        Args:
            in_data (obj): Data to read.
        """
        # if the data is already set, let's subtract it from the length before adding the new data
        if self.cert_data is not None:
            self.Hdr_dwLength -= len(self.cert_data)

        # Account for back compat. Behave differently for file streams.
        if hasattr(in_data, 'seek'):
            self.cert_data = in_data.read()
        elif hasattr(in_data, 'decode'):
            self.cert_data = in_data
        else:
            raise ValueError(
                f"Invalid datatype provided: {type(in_data)}, data may only be of type filestream or bytes")

        self.Hdr_dwLength = self.Hdr_dwLength + len(self.cert_data)

    def PopulateFromFileStream(self, fs):
        """Un-serialized from a filestream.

        Args:
            fs (obj): Already opened file

        Raises:
            (ValueError): Invalid fs
            (ValueError): Invalid size
        """
        warn("PopulateFromFileStream is deprecated, use decode instead", DeprecationWarning, 2)
        return self.decode(fs)

    def get_certificate(self):
        """Returns certificate data, if certificate data exists."""
        return self.cert_data

    def GetCertificate(self):
        """Returns certificate data, if certificate data exists."""
        warn("GetCertificate is deprecated, use get_certificate instead", DeprecationWarning, 2)
        return self.get_certificate()

    def print(self, outfs=sys.stdout):
        """Prints struct to console."""
        self.dump_info(outfs)

    def Print(self, outfs=sys.stdout):
        """Prints struct to console."""
        warn("Print() is deprecated, use print() instead.", DeprecationWarning, 2)
        self.dump_info(outfs)

    def dump_info(self, outfs=sys.stdout):
        """Prints struct to a file stream."""
        outfs.write("\n-------------------- WIN_CERTIFICATE ---------------------\n")
        outfs.write(f"WIN_CERTIFICATE.dwLength         = {self.Hdr_dwLength:08X}\n")
        outfs.write(f"WIN_CERTIFICATE.wRevision        = {self.Hdr_wRevision:04X}\n")
        outfs.write(f"WIN_CERTIFICATE.wCertificateType = {self.Hdr_wCertificateType:04X}\n")
        outfs.write(f"WIN_CERTIFICATE_UEFI_GUID.cert_type             = {str(self.cert_type).upper()}\n")
        outfs.write(f"sizeof (WIN_CERTIFICATE_UEFI_GUID.cert_data)    = {len(self.cert_data):08X}\n")

        outfs.write("\n------------------- CERTIFICATE DATA ---------------------\n")

        # Technically the signature could be wrapped in a ContentInfo
        try:
            content_info, _ = der_decode(self.cert_data, asn1Spec=rfc2315.ContentInfo())

            outfs.write(str(content_info))
        except Exception:
            # But usually its not
            try:
                signed_data, _ = der_decode(self.cert_data, asn1Spec=rfc2315.SignedData())

                outfs.write(str(signed_data))
            except Exception as exc:
                raise ValueError("Unable to decode cert_data") from exc

    def DumpInfo(self, outfs=sys.stdout):
        """Prints struct to a file stream."""
        warn("DumpInfo is deprecated, use dump_info instead", DeprecationWarning, 2)
        self.dump_info(outfs)

    def write(self, fs):
        """Writes the struct to a filestream."""
        fs.write(self.encode())

    def Write(self, fs):
        """Writes the struct to a filestream."""
        warn("Write is deprecated, use write instead", DeprecationWarning, 2)
        fs.write(self.encode())

    def __str__(self):
        """Returns the object as a string."""
        string_repr = ""
        with io.StringIO() as f:
            self.print(f)
            f.seek(0)
            string_repr = f.read()

        return string_repr


class WinCert(object):
    """Object for generating a WinCert."""
    STATIC_STRUCT_SIZE = 8
    # WIN_CERTIFICATE.wCertificateTypes UEFI Spec defined
    WIN_CERT_TYPE_NONE = 0x0000
    WIN_CERT_TYPE_PKCS_SIGNED_DATA = 0x0002
    WIN_CERT_TYPE_EFI_PKCS115 = 0x0EF0
    WIN_CERT_TYPE_EFI_GUID = 0x0EF1
    # Revision
    REVISION = 0x200

    @staticmethod
    def factory(fs):
        """Generates a specific Cert Type depending on parsed Hdr_wCertificationType from the fs.

        Args:
            fs (obj): filestream

        Returns:
            (WinCertUefiGuid): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_GUID
            (WinCertPkcs1): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_PKCS115
        Raises:
            (Exception): Invalid fs
            (Exception): Invalid fs size
        """
        if (fs is None):
            raise Exception("Invalid File stream")

        # only populate from file stream those parts that are complete in the file stream
        offset = fs.tell()
        fs.seek(0, 2)
        end = fs.tell()
        fs.seek(offset)

        if ((end - offset) < WinCert.STATIC_STRUCT_SIZE):  # size of the static header data
            raise Exception("Invalid file stream size")
        # 1 read len
        # 2 read revision
        # 3 read cert type
        fs.seek(4, 1)  # seeking past Hdr_dwLength
        fs.seek(2, 1)  # seeking past Hdr_wRevision
        Hdr_wCertificateType = struct.unpack("=H", fs.read(2))[0]

        fs.seek(offset)

        if (Hdr_wCertificateType == WinCert.WIN_CERT_TYPE_EFI_GUID):
            return WinCertUefiGuid(fs)
        elif (Hdr_wCertificateType == WinCert.WIN_CERT_TYPE_EFI_PKCS115):
            return WinCertPkcs1(fs)
        else:
            return None
    #
    # this method is a factory
    #

    @staticmethod
    def Factory(fs):
        """Generates a specific Cert Type depending on parsed Hdr_wCertificationType from the fs.

        Args:
            fs (obj): filestream

        Returns:
            (WinCertUefiGuid): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_GUID
            (WinCertPkcs1): if Hdr_wCertificationType == WIN_CERT_TYPE_EFI_PKCS115
        Raises:
            (Exception): Invalid fs
            (Exception): Invalid fs size
        """
        warn("Factory is deprecated, use factory instead", DeprecationWarning, 2)
        return WinCert.factory(fs)



##
# Copyright (C) Microsoft Corporation. All rights reserved.
# SPDX-License-Identifier: BSD-2-Clause-Patent
#
# Python script that converts a raw IVRS table into a struct, the spec version is based on
# https://www.amd.com/system/files/TechDocs/48882_IOMMU.pdf
##
"""Module for converting a raw IVRS table into a struct."""

import struct
import sys
import xml.etree.ElementTree as ET
from enum import IntEnum

IVRSParserVersion = '1.00'
# spell-checker:ignore IVMD, IOMMUEFR


class IVRS_TABLE(object):
    """Object representing a IVRS Table."""
    def __init__(self, data=None):
        """Inits an empty object."""
        self.acpi_header = None
        self.SubStructs = list()
        self.IVMD_list = list()

        if data is not None:
            self.Decode(data)

    def Decode(self, data):
        """Loads data into the object from the header_byte_array."""
        self.acpi_header = IVRS_TABLE.ACPI_TABLE_HEADER(data[:IVRS_TABLE.ACPI_TABLE_HEADER.struct_format_size])

        # Start from the end of ACPI header, but store the parsed length for verification
        t_length = self.acpi_header.Length
        self.acpi_header.Length = IVRS_TABLE.ACPI_TABLE_HEADER.struct_format_size
        t_data = data[IVRS_TABLE.ACPI_TABLE_HEADER.struct_format_size:]

        # sanity check on incoming data
        Checksum8 = IVRS_TABLE.validateChecksum8(data)
        if (Checksum8 != 0):
            raise Exception('Incoming data checksum does not add up: checksum field %x, calculated is %x' %
                            (self.acpi_header.Checksum, Checksum8))

        while len(t_data) > 0:
            # Get type and length of remapping struct
            remapping_header = self.REMAPPING_STRUCT_HEADER(t_data)

            # Parse remapping struct
            if (remapping_header.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_10H) or\
               (remapping_header.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_11H) or\
               (remapping_header.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_40H):
                remapping_header = self.IVHD_STRUCT(t_data)
                self.addIVHDEntry(remapping_header)
            elif (remapping_header.Type == IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_20H) or\
                 (remapping_header.Type == IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_21H) or\
                 (remapping_header.Type == IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_22H):
                remapping_header = self.IVMD_STRUCT(t_data[:IVRS_TABLE.IVMD_STRUCT.struct_format_size])
                self.addIVMDEntry(remapping_header)
                if (remapping_header.Type == IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_20H):
                    self.IVRSBit = 0
            else:
                print('Reserved remapping struct found in IVRS table %d' % remapping_header.Type)
                sys.exit(-1)

            # Update data position
            t_data = t_data[remapping_header.Length:]

        if (self.acpi_header.Length != t_length) or (len(t_data) != 0):
            raise Exception("IVRS length does not add up. Parsed len: %d, reported len: %d" %
                            (t_length, self.acpi_header.Length))

    def Encode(self):
        r"""Serializes the object.

        Returns:
            (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
        """
        bytes_str = b''

        # Append ACPI header
        bytes_str += self.acpi_header.Encode()

        # All IVHD/IVMD entries
        for ivxd in self.SubStructs:
            bytes_str += ivxd.Encode()
        return bytes_str

    def ToXmlElementTree(self):
        """Transforms the object into an xml item and returns it."""
        root = ET.Element('IVRSTable')
        root.append(self.acpi_header.ToXmlElementTree())
        for sub in self.SubStructs:
            root.append(sub.ToXmlElementTree())

        return root

    def DumpInfo(self):
        """Prints internal information to the console."""
        self.acpi_header.DumpInfo()

        for sub in self.SubStructs:
            sub.DumpInfo()

    @staticmethod
    def validateChecksum8(data):
        """Validate the Checksum."""
        return sum(data) & 0xFF

    def updateACPISum(self):
        """Update the ACPI Checksum."""
        temp_sum = 0
        # Clear the checksum before calculating sum
        self.acpi_header.Checksum = 0
        temp_str = self.Encode()
        temp_sum = sum(temp_str)
        self.acpi_header.Checksum = (0x100 - (temp_sum & 0xFF)) & 0xFF

    def addIVHDEntry(self, ivhd):
        """Append entry to the list, update length and checksum."""
        self.acpi_header.Length += len(ivhd.Encode())
        self.SubStructs.append(ivhd)
        self.updateACPISum()

    def addIVMDEntry(self, ivmd):
        """Append entry to the list, update length and checksum."""
        self.acpi_header.Length += len(ivmd.Encode())
        # IVMD has to follow the corresponding IVHD, thus the list records all entries to maintain order
        self.SubStructs.append(ivmd)
        self.IVMD_list.append(ivmd)
        self.updateACPISum()

    def IVRSBitEnabled(self):
        """Returns the IVRSBit Value."""
        return bool(self.acpi_header.IVRSBit)

    class ACPI_TABLE_HEADER(object):
        """Object representing a ACPI Table Header."""
        struct_format = '=4sIBB6s8sI4sIIQ'
        struct_format_size = struct.calcsize(struct_format)

        def __init__(self, data=None):
            """Inits an empty object."""
            self.Signature = None
            self.Length = 0
            self.Revision = 0
            self.Checksum = 0
            self.OEMID = 0
            self.OEMTableID = 0
            self.OEMRevision = 0
            self.CreatorID = 0
            self.CreatorRevision = 0
            self.IVinfo = None
            self.Reserved = 0

            self.IVRSBit = 0

            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Signature,
             self.Length,
             self.Revision,
             self.Checksum,
             self.OEMID,
             self.OEMTableID,
             self.OEMRevision,
             self.CreatorID,
             self.CreatorRevision,
             self.IVinfo,
             self.Reserved) = struct.unpack(IVRS_TABLE.ACPI_TABLE_HEADER.struct_format, header_byte_array)

            self.IVRSBit = self.IVinfo & 0x02
            if (self.IVinfo & 0x1E) == 0:
                sys.exit(-1)

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(self.struct_format,
                               self.Signature,
                               self.Length,
                               self.Revision,
                               self.Checksum,
                               self.OEMID,
                               self.OEMTableID,
                               self.OEMRevision,
                               self.CreatorID,
                               self.CreatorRevision,
                               self.IVinfo,
                               self.Reserved)

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('  ACPI Table Header')
            print('------------------------------------------------------------------')
            print('Signature          : {Signature:s}'.format(Signature=self.Signature.decode()))
            print('Length             : 0x{Length:08X}'.format(Length=self.Length))
            print('Revision           : 0x{Revision:02X}'.format(Revision=self.Revision))
            print('Checksum           : 0x{Checksum:02X}'.format(Checksum=self.Checksum))
            print('OEM ID             : {OEMID:s}'.format(OEMID=self.OEMID.decode()))
            print('OEM Table ID       : {OEMTableID:s}'.format(OEMTableID=self.OEMTableID.decode()))
            print('OEM Revision       : 0x{OEMRevision:08X}'.format(OEMRevision=self.OEMRevision))
            print('Creator ID         : {CreatorID:s}'.format(CreatorID=self.CreatorID.decode()))
            print('Creator Revision   : 0x{CreatorRevision:08X}'.format(CreatorRevision=self.CreatorRevision))
            print('IVinfo             : 0x{IVinfo:08X}'.format(IVinfo=self.IVinfo))

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_repr = ET.Element('AcpiTableHeader')
            xml_repr.set('Signature', '%s' % self.Signature)
            xml_repr.set('Length', '0x%X' % self.Length)
            xml_repr.set('Revision', '0x%X' % self.Revision)
            xml_repr.set('Checksum', '0x%X' % self.Checksum)
            xml_repr.set('OEMID', '%s' % self.OEMID)
            xml_repr.set('OEMTableID', '%s' % self.OEMTableID)
            xml_repr.set('OEMRevision', '0x%X' % self.OEMRevision)
            xml_repr.set('CreatorID', '%s' % self.CreatorID)
            xml_repr.set('CreatorRevision', '0x%X' % self.CreatorRevision)
            xml_repr.set('IVinfo', '0x%X' % self.IVinfo)
            return xml_repr

    class REMAPPING_STRUCT_HEADER(object):
        """Generic ReMapping Struct Header."""
        struct_format = '=B'
        struct_format_size = struct.calcsize(struct_format)

        def __init__(self, header_byte_array):
            """Inits an empty object."""
            (self.Type, ) = struct.unpack(IVRS_TABLE.REMAPPING_STRUCT_HEADER.struct_format,
                                          header_byte_array[:IVRS_TABLE.REMAPPING_STRUCT_HEADER.struct_format_size])

    class IVHD_STRUCT(REMAPPING_STRUCT_HEADER):
        """Object representing a IVHD Struct."""
        # cspell:disable-next disable the spell checker from thinking this a word
        struct_format = '=BBHHHQHHI'
        struct_format_size = struct.calcsize(struct_format)
        ex_format = "=QQ"
        ex_format_size = struct.calcsize(ex_format)

        class IVHD_TYPE(IntEnum):
            """IVMD_STRUCT type enum."""
            TYPE_10H = 0x10
            TYPE_11H = 0x11
            TYPE_40H = 0x40

        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = None
            self.Flags = None
            self.Length = 0
            self.DeviceID = 0
            self.CapabilityOffset = 0
            self.IOMMUBaseAddress = 0
            self.SegmentGroup = 0
            self.IOMMUInfo = None
            self.IOMMUFeatureInfo = None
            self.IOMMUEFRImage = None
            self.Reserved = 0

            self.DeviceTableEntries = list()

            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.Flags,
             t_Length,
             self.DeviceID,
             self.CapabilityOffset,
             self.IOMMUBaseAddress,
             self.SegmentGroup,
             self.IOMMUInfo,
             self.IOMMUFeatureInfo) = struct.unpack(IVRS_TABLE.IVHD_STRUCT.struct_format,
                                                    header_byte_array[:IVRS_TABLE.IVHD_STRUCT.struct_format_size])

            self.Length = 0

            if (self.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_11H) or\
               (self.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_40H):
                ivhd_hdr_size = (IVRS_TABLE.IVHD_STRUCT.struct_format_size + IVRS_TABLE.IVHD_STRUCT.ex_format_size)
                (self.IOMMUEFRImage, self.Reserved) =\
                    struct.unpack(IVRS_TABLE.IVHD_STRUCT.ex_format,
                                  header_byte_array[IVRS_TABLE.IVHD_STRUCT.struct_format_size:ivhd_hdr_size])
            else:
                ivhd_hdr_size = IVRS_TABLE.IVHD_STRUCT.struct_format_size

            header_byte_array = header_byte_array[ivhd_hdr_size:]
            bytes_left = t_Length - ivhd_hdr_size
            self.Length += ivhd_hdr_size

            # Get Sub Structs
            while bytes_left > 0:
                device_scope = IVRS_TABLE.DEVICE_TABLE_ENTRY.Factory(header_byte_array)
                header_byte_array = header_byte_array[device_scope.Length:]
                bytes_left -= device_scope.Length
                if (device_scope.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END):
                    self.addDTEEntry(device_scope)

            if (t_Length != self.Length) or (bytes_left != 0):
                raise Exception("IVHD length does not add up. Parsed len: %d, reported len: %d" %
                                (self.Length, t_Length))

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = b''
            byte_str += struct.pack(IVRS_TABLE.IVHD_STRUCT.struct_format,
                                    self.Type,
                                    self.Flags,
                                    self.Length,
                                    self.DeviceID,
                                    self.CapabilityOffset,
                                    self.IOMMUBaseAddress,
                                    self.SegmentGroup,
                                    self.IOMMUInfo,
                                    self.IOMMUFeatureInfo)

            if self.IOMMUEFRImage is not None:
                byte_str += struct.pack(IVRS_TABLE.IVHD_STRUCT.ex_format, self.IOMMUEFRImage, self.Reserved)

            for dte in self.DeviceTableEntries:
                byte_str += dte.Encode()

            return byte_str

        def addDTEEntry(self, dte):
            """Append raw data, update length. checksum will be left untouched."""
            self.Length += len(dte.Encode())
            self.DeviceTableEntries.append(dte)

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_repr = ET.Element('IVHD')
            xml_repr.set('Type', '0x%X' % self.Type)
            xml_repr.set('Flags', '0x%X' % self.Flags)
            xml_repr.set('Length', '0x%X' % self.Length)
            xml_repr.set('IOMMUDeviceID', '0x%X' % self.DeviceID)
            xml_repr.set('CapabilityOffset', '0x%X' % self.CapabilityOffset)
            xml_repr.set('IOMMUBaseAddress', '0x%X' % self.IOMMUBaseAddress)
            xml_repr.set('SegmentGroup', '0x%X' % self.SegmentGroup)
            xml_repr.set('IOMMUInfo', '0x%X' % self.IOMMUInfo)
            xml_repr.set('IOMMUFeatureInfo', '0x%X' % self.IOMMUFeatureInfo)

            if (self.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_11H) or\
               (self.Type == IVRS_TABLE.IVHD_STRUCT.IVHD_TYPE.TYPE_40H):
                xml_repr.set('IOMMUEFRImage', '0x%X' % self.IOMMUEFRImage)

            # Add SubStructs
            for item in self.DeviceTableEntries:
                xml_repr.append(item.ToXmlElementTree())

            return xml_repr

        def DumpInfo(self):
            """Prints internal information to the console."""
            print("\t  IVHD")
            print("\t----------------------------------------------------------------")
            print('\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\tFlags                 : 0x{Flags:02X}'.format(Flags=self.Flags))
            print('\tLength                : 0x{Length:04X}'.format(Length=self.Length))
            print('\tIOMMU Device ID       : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\tCapability Offset     : 0x{CapabilityOffset:04X}'.format(CapabilityOffset=self.CapabilityOffset))
            print('\tIOMMU Base Address    : 0x{IOMMUBaseAddress:016X}'.format(IOMMUBaseAddress=self.IOMMUBaseAddress))
            print('\tSegment Group         : 0x{SegmentGroup:04X}'.format(SegmentGroup=self.SegmentGroup))
            print('\tIOMMU Info            : 0x{IOMMUInfo:04X}'.format(IOMMUInfo=self.IOMMUInfo))
            print('\tIOMMU Feature Info    : 0x{IOMMUFeatureInfo:08X}'.format(IOMMUFeatureInfo=self.IOMMUFeatureInfo))

            for item in self.DeviceTableEntries:
                item.DumpInfo()

    class IVMD_STRUCT(REMAPPING_STRUCT_HEADER):
        """Object representing an IVMD Struct."""
        # cspell:disable-next disable spell checker from thinking this is a word
        struct_format = '=BBHHHQQQ'
        struct_format_size = struct.calcsize(struct_format)

        class IVMD_TYPE(IntEnum):
            """IVMD_STRUCT type enum."""
            TYPE_20H = 0x20  # All peripherals
            TYPE_21H = 0x21  # Specified peripheral
            TYPE_22H = 0x22  # Peripheral range

        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = None
            self.Flags = None
            self.Length = 0
            self.DeviceID = 0
            self.AuxiliaryData = None
            self.Reserved = 0
            self.IVMDStartAddress = 0
            self.IVMDMemoryBlockLength = 0

            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.Flags,
             self.Length,
             self.DeviceID,
             self.AuxiliaryData,
             self.Reserved,
             self.IVMDStartAddress,
             self.IVMDMemoryBlockLength) = struct.unpack(IVRS_TABLE.IVMD_STRUCT.struct_format, header_byte_array)
            # IVMD is simple, the length is fixed, so assert if not
            if (self.Length != len(header_byte_array)):
                raise Exception("Bad IVMD entry size %d, expecting %d" % (self.Length, len(header_byte_array)))

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(IVRS_TABLE.IVMD_STRUCT.struct_format,
                               self.Type,
                               self.Flags,
                               self.Length,
                               self.DeviceID,
                               self.AuxiliaryData,
                               self.Reserved,
                               self.IVMDStartAddress,
                               self.IVMDMemoryBlockLength)

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_repr = ET.Element('IVMD')

            xml_repr.set('Type', '0x%X' % self.Type)
            xml_repr.set('Flags', '0x%X' % self.Flags)
            xml_repr.set('Length', '0x%X' % self.Length)
            xml_repr.set('DeviceID', '0x%X' % self.DeviceID)
            if (self.Type != IVRS_TABLE.IVMD_STRUCT.IVMD_TYPE.TYPE_22H):
                xml_repr.set('AuxiliaryData', '0x%X' % self.AuxiliaryData)
            else:
                xml_repr.set('EndofRange', '0x%X' % self.AuxiliaryData)
            xml_repr.set('Reserved', '0x%X' % self.Reserved)
            xml_repr.set('IVMDStartAddress', '0x%X' % self.IVMDStartAddress)
            xml_repr.set('IVMDMemoryBlockLength', '0x%X' % self.IVMDMemoryBlockLength)

            return xml_repr

        def DumpInfo(self):
            """Prints internal information to the console."""
            print("\t  IVMD")
            print("\t----------------------------------------------------------------")
            print('\tType                                 : 0x{Type:02X}'.format(Type=self.Type))
            print('\tFlags                                : 0x{Flags:02X}'.format(Flags=self.Flags))
            print('\tLength                               : 0x{Length:04X}'.format(Length=self.Length))
            print('\tDeviceID                             : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\tAuxiliaryData                        : 0x{AuxiliaryData:04X}'.
                  format(AuxiliaryData=self.AuxiliaryData))
            print('\tReserved                             : 0x{Reserved:016X}'.format(Reserved=self.Reserved))
            print('\tIVMD Start Address                   : 0x{IVMDStartAddress:016X}'.
                  format(IVMDStartAddress=self.IVMDStartAddress))
            print('\tIVMD Memory Block Length             : 0x{IVMDMemoryBlockLength:016X}'.
                  format(IVMDMemoryBlockLength=self.Type))

    class DEVICE_TABLE_ENTRY(object):
        """A Generic Device Table Entry."""
        struct_format = '=BHB'
        struct_format_size = struct.calcsize(struct_format)
        dte_var_ext_format = "=8s8sBB"
        dte_var_len = struct_format_size + struct.calcsize(dte_var_ext_format)

        class DTE_TYPE(IntEnum):
            """Enum for Device Table Entry Types."""
            RESERVED = 0
            ALL = 1
            SELECT = 2
            RANGE_START = 3
            RANGE_END = 4
            ALIAS_SELECT = 66
            ALIAS_RANGE_START = 67
            EX_SELECT = 70
            EX_RANGE_START = 71
            SPECIAL = 72
            ACPI = 240

        @staticmethod
        def Factory(data):
            """Factory method to generate a specific table entry type."""
            if (data is None):
                raise Exception("Invalid File stream")

            RemapHeader = IVRS_TABLE.REMAPPING_STRUCT_HEADER(data)
            Type = RemapHeader.Type

            if (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RESERVED):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_RESERVED(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALL):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_ALL(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SELECT):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_SELECT(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_START):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_RANGE_START(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_SELECT):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_ALIAS_SELECT(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_RANGE_START):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_ALIAS_RANGE_START(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_SELECT):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_EX_SELECT(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_RANGE_START):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_EX_RANGE_START(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SPECIAL):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_SPECIAL(data)
            elif (Type == IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ACPI):
                return IVRS_TABLE.DEVICE_TABLE_ENTRY_ACPI(data)
            else:
                return None

    class DEVICE_TABLE_ENTRY_RESERVED(object):
        """Object representing a Device Table Entry RESERVED."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RESERVED):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RESERVED, self.Type)

            self.TypeString = "Reserved"
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                               self.Type,
                               self.DeviceID,
                               self.DTESetting)

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('DeviceID', '0x%X' % (self.DeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            xml_item.set('Type', '0x%X' % self.Type)
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tDevice ID             : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))

    class DEVICE_TABLE_ENTRY_ALL(object):
        """Object representing a Device Table Entry ALL."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALL):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALL,
                                self.Type)

            self.TypeString = "All"
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                               self.Type,
                               self.DeviceID,
                               self.DTESetting)

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('DeviceID', '0x%X' % (self.DeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tDevice ID             : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))

    class DEVICE_TABLE_ENTRY_SELECT(object):
        """Object representing a Device Table Entry SELECT."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SELECT):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SELECT,
                                self.Type)

            self.TypeString = "Reserved"
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            return struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                               self.Type,
                               self.DeviceID,
                               self.DTESetting)

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('DeviceID', '0x%X' % (self.DeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tDevice ID             : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))

    class DEVICE_TABLE_ENTRY_RANGE_START(object):
        """Object representing a Device Table Entry RANGE_START."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_START):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_START,
                                self.Type)

            self.TypeString = "Range"
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            (Type, self.EndDeviceID, _) =\
                struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                              header_byte_array[self.Length:
                                                (self.Length + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size)])
            if Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END:
                print("Start of range does not follow end of range")
                sys.exit(-1)
            self.Length += IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                   self.Type,
                                   self.DeviceID,
                                   self.DTESetting)
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END,
                                    self.EndDeviceID,
                                    0)
            return byte_str

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('StartofRange', '0x%X' % self.DeviceID)
            xml_item.set('EndofRange', '0x%X' % (self.EndDeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tStart of Range        : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tEnd of Range          : 0x{EndDeviceID:04X}'.format(EndDeviceID=self.EndDeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))

    class DEVICE_TABLE_ENTRY_ALIAS_SELECT(object):
        """Object representing a Device Table Entry ALIAS_SELECT."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.SourceDeviceID = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_SELECT):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_SELECT,
                                self.Type)

            self.TypeString = "Alias Select"
            # Two DevID, one for alias, one for source
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size +\
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            (_, self.SourceDeviceID, _) =\
                struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                              header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size:self.Length])

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                   self.Type,
                                   self.DeviceID,
                                   self.DTESetting)
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, 0, self.SourceDeviceID, 0)
            return byte_str

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('DeviceID', '0x%X' % (self.DeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            xml_item.set('SourceDeviceID', '0x%X' % (self.SourceDeviceID))
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tDevice ID             : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))
            print('\t\tSource Device ID      : 0x{SourceDeviceID:04X}'.format(SourceDeviceID=self.SourceDeviceID))

    class DEVICE_TABLE_ENTRY_ALIAS_RANGE_START(object):
        """Object representing a Device Table Range Start."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.SourceDeviceID = 0
            self.EndDeviceID = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_RANGE_START):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ALIAS_RANGE_START,
                                self.Type)

            self.TypeString = "Alias Range"
            # Two DevID, one for alias start, one for source start
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size +\
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            (_, self.SourceDeviceID, _) =\
                struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                              header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size:self.Length])
            (Type, self.EndDeviceID, _) =\
                struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                              header_byte_array[self.Length:
                                                (self.Length + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size)])
            if Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END:
                print("Start of range does not follow end of range")
                sys.exit(-1)
            self.Length += IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                   self.Type,
                                   self.DeviceID,
                                   self.DTESetting)
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, 0, self.SourceDeviceID, 0)
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                    IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END,
                                    self.EndDeviceID,
                                    0)
            return byte_str

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('StartofRange', '0x%X' % self.DeviceID)
            xml_item.set('EndofRange', '0x%X' % (self.EndDeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            xml_item.set('SourceDeviceID', '0x%X' % (self.SourceDeviceID))
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tStart of Range        : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tEnd of Range          : 0x{EndDeviceID:04X}'.format(EndDeviceID=self.EndDeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))
            print('\t\tSource Device ID      : 0x{SourceDeviceID:04X}'.format(SourceDeviceID=self.SourceDeviceID))

    class DEVICE_TABLE_ENTRY_EX_SELECT(object):
        """Object representing a Device Table Entry EX_SELECT."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.ExtendedDTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_SELECT):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_SELECT,
                                self.Type)

            self.TypeString = "Extended Select"
            # Two DTE setting, one for standard setting, one for extended setting (AtsDisabled, etc.)
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size +\
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            (self.ExtendedDTESetting,) = \
                struct.unpack("=I", header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size:self.Length])

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                   self.Type,
                                   self.DeviceID,
                                   self.DTESetting)
            # Two DTE setting, one for standard setting, one for extended setting (AtsDisabled, etc.)
            byte_str += struct.pack("=I", self.ExtendedDTESetting)
            return byte_str

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('DeviceID', '0x%X' % (self.DeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            if (self.ExtendedDTESetting & 0x80000000) != 0:
                xml_item.set('ExtendedDTESetting', 'ATS requests blocked')
            else:
                xml_item.set('ExtendedDTESetting', 'ATS allowed')
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tDevice ID             : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))
            if (self.ExtendedDTESetting & 0x80000000) != 0:
                ats_str = "ATS requests blocked"
            else:
                ats_str = "ATS allowed"
            print('\t\tExtended DTE Setting  : {ExtendedDTESetting:s}'.format(ExtendedDTESetting=ats_str))

    class DEVICE_TABLE_ENTRY_EX_RANGE_START(object):
        """Object representing a Device Table Entry EX_RANGE_START."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_RANGE_START):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.EX_RANGE_START,
                                self.Type)

            self.TypeString = "Extended Range"
            # Two DTE setting, one for standard setting start, one for extended setting start
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size +\
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            (self.ExtendedDTESetting,) =\
                struct.unpack("=I", header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size:self.Length])
            (Type, self.EndDeviceID, _) =\
                struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                              header_byte_array[self.Length:
                                                (self.Length + IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size)])
            if Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.RANGE_END:
                print("Start of range does not follow end of range")
                sys.exit(-1)
            self.Length += IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                   self.Type,
                                   self.DeviceID,
                                   self.DTESetting)
            # Two DTE setting, one for standard setting start, one for extended setting start
            byte_str += struct.pack("=I", self.ExtendedDTESetting)
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format, 4, self.EndDeviceID, 0)
            return byte_str

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('StartofRange', '0x%X' % self.DeviceID)
            xml_item.set('EndofRange', '0x%X' % (self.EndDeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            if (self.ExtendedDTESetting & 0x80000000) != 0:
                xml_item.set('ExtendedDTESetting', 'ATS requests blocked')
            else:
                xml_item.set('ExtendedDTESetting', 'ATS allowed')
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tStart of Range        : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tEnd of Range          : 0x{EndDeviceID:04X}'.format(EndDeviceID=self.EndDeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))
            if (self.ExtendedDTESetting & 0x80000000) != 0:
                ats_str = "ATS requests blocked"
            else:
                ats_str = "ATS allowed"
            print('\t\tExtended DTE Setting  : {ExtendedDTESetting:s}'.format(ExtendedDTESetting=ats_str))

    class DEVICE_TABLE_ENTRY_SPECIAL(object):
        """Object representing a Device Table Entry Special."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.ExtendedDTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SPECIAL):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.SPECIAL,
                                self.Type)

            self.TypeString = "Special Device"
            # First half for standard DTE setting, second half for special DevID and its variety (APIC, HPET, etc.)
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size +\
                IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size
            (self.Handle, self.SourceDeviceID, self.Variety) =\
                struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                              header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size:self.Length])

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                   self.Type,
                                   self.DeviceID,
                                   self.DTESetting)
            # First half for standard DTE setting, second half for special DevID and its variety (APIC, HPET, etc.)
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                    self.Handle,
                                    self.SourceDeviceID,
                                    self.Variety)
            return byte_str

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('DeviceID', '0x%X' % (self.DeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))
            xml_item.set('SourceDeviceID', '0x%X' % (self.SourceDeviceID))

            xml_item.set('Handle', '0x%X' % (self.Handle))
            if self.Variety == 1:
                xml_item.set('Variety', 'IOAPIC')
            elif self.Variety == 2:
                xml_item.set('Variety', 'HPET')
            else:
                xml_item.set('Variety', 'Reserved %X' % (self.Variety))
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tDevice ID             : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))
            print('\t\tSource Device ID      : 0x{SourceDeviceID:04X}'.format(SourceDeviceID=self.SourceDeviceID))

            if self.Variety == 1:
                var_str = "IOAPIC"
            elif self.Variety == 2:
                var_str = "HPET"
            else:
                var_str = "Reserved 0x%02X" % (self.Variety)
            print('\t\tHandle                : 0x{Handle:02X}'.format(Handle=self.Handle))
            print('\t\tVariety               : {Variety:s}'.format(Variety=var_str))

    class DEVICE_TABLE_ENTRY_ACPI(object):
        """Object representing a Device Table Entry ACPI."""
        def __init__(self, data=None):
            """Inits an empty object."""
            self.Type = 0
            self.DeviceID = 0
            self.DTESetting = 0
            self.ExtendedDTESetting = 0
            self.TypeString = None
            self.Length = 0
            if data is not None:
                self.Decode(data)

        def Decode(self, header_byte_array):
            """Loads data into the object from the header_byte_array."""
            (self.Type,
             self.DeviceID,
             self.DTESetting) = struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                              header_byte_array[:IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size])
            if (self.Type != IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ACPI):
                raise Exception("Input device type (%d) does not match expectation (%d)",
                                IVRS_TABLE.DEVICE_TABLE_ENTRY.DTE_TYPE.ACPI,
                                self.Type)

            self.TypeString = "Variable Length ACPI HID Device"
            (self.HID, self.CID, self.UIDFormat, self.UIDLength) =\
                struct.unpack(IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_ext_format,
                              header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format_size:
                                                IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_len])
            self.Length = IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_len + self.UIDLength
            if self.UIDFormat == 0:
                self.UID = None
            elif self.UIDFormat == 1:
                (self.UID,) = struct.unpack("=Q", header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_len:
                                                                    self.Length])
            elif self.UIDFormat == 2:
                (self.UID,) =\
                    struct.unpack("=%ss" % self.UIDLength,
                                  header_byte_array[IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_len:self.Length])

        def Encode(self):
            r"""Serializes the object.

            Returns:
                (str): string representing packed data as bytes (i.e. b'\x01\x00\x03')
            """
            byte_str = struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.struct_format,
                                   self.Type,
                                   self.DeviceID,
                                   self.DTESetting)
            # Variable Length ACPI HID Device
            byte_str += struct.pack(IVRS_TABLE.DEVICE_TABLE_ENTRY.dte_var_ext_format,
                                    self.HID,
                                    self.CID,
                                    self.UIDFormat,
                                    self.UIDLength)
            if self.UIDFormat == 1:
                byte_str += struct.pack("=Q", self.UID)
            elif self.UIDFormat == 2:
                byte_str += struct.pack("=%ss" % self.UIDLength, self.UID)
            return byte_str

        def ToXmlElementTree(self):
            """Transforms the object into an xml item and returns it."""
            xml_item = ET.Element(self.TypeString.replace(" ", ""))
            xml_item.set('Type', '0x%X' % self.Type)
            xml_item.set('DeviceID', '0x%X' % (self.DeviceID))
            xml_item.set('DTESetting', '0x%X' % (self.DTESetting))

            xml_item.set('HardwareID', '%s' % (self.HID))
            xml_item.set('ExtendedDTE Setting', '%s' % (self.CID))
            xml_item.set('UniqueIDFormat', '%d' % (self.UIDFormat))
            xml_item.set('UniqueIDLength', '%d' % (self.UIDLength))
            if self.UIDFormat == 0:
                xml_item.set('UniqueID', 'None')
            elif self.UIDFormat == 1:
                xml_item.set('UniqueID', '0x%X' % (self.UID))
            elif self.UIDFormat == 2:
                xml_item.set('UniqueID', '%s' % (self.UID))
            else:
                print("Unrecognized UID format detected")
                sys.exit(-1)
            return xml_item

        def DumpInfo(self):
            """Prints internal information to the console."""
            print('\t\t  {TypeString:s}'.format(TypeString=self.TypeString))
            print('\t\t--------------------------------------------------')
            print('\t\tType                  : 0x{Type:02X}'.format(Type=self.Type))
            print('\t\tDevice ID             : 0x{DeviceID:04X}'.format(DeviceID=self.DeviceID))
            print('\t\tDTE Setting           : 0x{DTESetting:02X}'.format(DTESetting=self.DTESetting))

            print('\t\tHardware ID           : {HID:s}'.format(HID=self.HID.decode()))
            print('\t\tExtended DTE Setting  : {CID:s}'.format(CID=self.CID.decode()))
            print('\t\tUnique ID Format      : {UIDFormat:d}'.format(UIDFormat=self.UIDFormat))
            print('\t\tUnique ID Length      : {UIDLength:d}'.format(UIDLength=self.UIDLength))
            if self.UIDFormat == 0:
                print('\t\tUnique ID             : None')
            elif self.UIDFormat == 1:
                print('\t\tUnique ID             : 0x{UID:X}'.format(UID=self.UID))
            elif self.UIDFormat == 2:
                print('\t\tUnique ID             : {UID:s}'.format(UID=self.UID.decode()))
            else:
                raise Exception("Unrecognized UID format detected %d" % self.UIDFormat)
